#include <string.h>
#include <stdlib.h>

#include "configure.h"

#include "settings.h"
#include "misc.h"
#include "pairs.h"
#include "str.h"


Pairs*
newPairs(int size, void* (*freeFunction)(void*))
{
	Pairs* pairs;
	//	if (!(pairs = CALLOC(sizeof(Pairs) + size * sizeof(Pair*)))) return NULL;
	//	pairs->pairs = (Pair**)(pairs + 1);
    size = biRound(size, 16);
	if (!(pairs = CALLOC(sizeof(Pairs)))) return NULL;
	pairs->pairs = CALLOC(size * sizeof(Pair*));
	pairs->size = size - 1;
	pairs->freeFunction = freeFunction;
	return pairs;
}


Pairs*
freePairs(Pairs* pairs)
{
	unsigned h;
	if (!pairs) return NULL;
	for (h = 0; h <= pairs->size; h++) {
		Pair* p = pairs->pairs[h];
		while (p) {
			Pair* next = p->next;
			if (pairs->freeFunction) pairs->freeFunction(p->value);
			FREE(p->name);
			FREE(p);
			p = next;
		}
	}
	FREE(pairs);
	return NULL;
}


Pair*
setPair(Pairs* pairs, char* name, void* value)
{
	unsigned hc;
	Pair* p;

	if (!value) {
		delPair(pairs, name);
		return NULL;
	}
	if (!(pairs && name)) return NULL;

	hc = hash(name) & pairs->size;
	p = pairs->pairs[hc];
	while(p) {
		if (!strcmp(p->name, name)) {
			if (pairs->freeFunction) pairs->freeFunction(p->value);
			p->value = value;
			return p;
		}
		p = p->next;
	}

	p = MALLOC(sizeof(Pair));
	p->name = STRDUP(name);
	p->value = value;
	p->next = pairs->pairs[hc];
	pairs->pairs[hc] = p;
	pairs->count++;
    return p;
}

void*
getPair(Pairs* pairs, char* name)
{
	Pair* p;
	if (!(pairs && name)) return NULL;
	p = pairs->pairs[hash(name) & pairs->size];
	while(p) {
		if (!strcmp(p->name, name)) return p->value;
		p = p->next;
	}
	return NULL;
}

Pair*
getPairPtr(Pairs* pairs, char* name)
{
	Pair* p;
	if (!(pairs && name)) return NULL;
	p = pairs->pairs[hash(name) & pairs->size];
	while(p) {
		if (!strcmp(p->name, name)) return p;
		p = p->next;
	}
	return NULL;
}

int
delPair(Pairs* pairs, char* name)
{
	Pair* p;
	Pair* prev = NULL;
	unsigned hc;

	if (!(pairs && name)) return FAILED;
    hc = hash(name) & pairs->size;
	p = pairs->pairs[hc];
	while(p) {
		if (!strcmp(p->name, name)) {
			if (pairs->freeFunction) pairs->freeFunction(p->value);
			if (prev) prev->next = p->next;
			else pairs->pairs[hc] = p->next;
            FREE(p->name);
            FREE(p);
			pairs->count--;
			return OK;
		}
        prev = p;
		p = p->next;
	}
	return FAILED;
	//return message(ERROR, "pair %s does not exists", name);;
}

int
renamePair(Pairs* pairs, char* name, char* new)
{
	Pair* p;
	Pair* prev = NULL;
	unsigned hc;

	if (!(pairs && name)) return FAILED;
    if (getPair(pairs, new)) return message(ERROR, "pair %s already exists", new);

	hc = hash(name) & pairs->size;
	p = pairs->pairs[hc];
	while(p) {
		if (!strcmp(p->name, name)) {
			if (prev) prev->next = p->next;
			else pairs->pairs[hc] = p->next;

			hc = hash(new) & pairs->size;
			p->next = pairs->pairs[hc];
			pairs->pairs[hc] = p;

			if (strlen(name) >= strlen(new))
				strcpy(p->name, new);
			else {
                FREE(p->name);
				p->name = STRDUP(new);
			}

			return OK;
		}
        prev = p;
		p = p->next;
	}
	return message(ERROR, "pair %s does not exists", name);;
}

void
startListPairs(Pairs* pairs)
{
	if (!pairs) return;
	pairs->last = NULL;
	pairs->lastHash = 0;
}

Pair*
nextPair(Pairs* pairs)
{
	if (pairs->last) {
		if (pairs->last->next) return pairs->last = pairs->last->next;
		++(pairs->lastHash);
	}

	for (; pairs->lastHash <= pairs->size; pairs->lastHash++)
		if (pairs->pairs[pairs->lastHash])
			return pairs->last = pairs->pairs[pairs->lastHash];

	pairs->lastHash = 0;
	pairs->last = NULL;

	return NULL;
}


static int
cmpPairByName(const void *f, const void *s)
{
    const Pair* p1 = *((const Pair**)f);
	const Pair* p2 = *((const Pair**)s);
	return strcmp(p1->name, p2->name);
}

static int
cmpPairByStrValue(const void *f, const void *s)
{
    const Pair* p1 = *((const Pair**)f);
	const Pair* p2 = *((const Pair**)s);
	return strcmp((char*)p1->value, (char*)p2->value);
}


Pair**
sortPairsByName(Pairs* pairs)
{
	Pair** sorted;
	unsigned h, n = 0;

	if (!pairs) return NULL;

	sorted = MALLOC(sizeof(Pair*) * pairs->count);

	for (h = 0; h <= pairs->size; h++) {
		Pair* p = pairs->pairs[h];
		while (p) {
            sorted[n++] = p;
			p = p->next;
		}
	}
	qsort(sorted, n, sizeof(Pair*), cmpPairByName);
    return sorted;
}

Pair**
sortPairsByStrValue(Pairs* pairs)
{
	Pair** sorted;
	unsigned h, n = 0;

	if (!pairs) return NULL;

	sorted = MALLOC(sizeof(Pair*) * pairs->count);

	for (h = 0; h <= pairs->size; h++) {
		Pair* p = pairs->pairs[h];
		while (p) {
			sorted[n++] = p;
			p = p->next;
		}
	}
	qsort(sorted, n, sizeof(Pair*), cmpPairByStrValue);
	return sorted;
}


