/* $Id: httpd.c,v 1.7 2008/02/17 04:42:11 maxim Exp $
 *
 */

#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/resource.h>
#include <sys/time.h>

#include "configure.h"

#include "common/misc.h"
#include "common/settings.h"
#include "common/signames.h"
#include "common/ip_ranges.h"
#include "common/str.h"
#if DEBUGLOG
#include "common/ptrinfo.h"
#endif
#include "config.h"
#include "core.h"
#include "httpd.h"



int doRestart, doReload, doShutdown, doQuit, doReExec, childExits;
time_t start, configured;

#if SUPPORT_COUNTRY
IPRanges* countries;
IPRanges* classes;
IPRanges* regions;
#endif

static char wrapper = NO;
static char master = YES;


void actionOnSignal(int sig);

void
reapChildren()
{
	int status, pid;
	signal(SIGCHLD, actionOnSignal);

	while (TRUE) {
		pid = waitpid(-1, &status, WNOHANG);
		if (pid <= 0) break;
		childExits = YES;

		if (WIFEXITED(status))
			message(MESSAGE, "child process %d exited with status %d", pid, WEXITSTATUS(status));
		else if (WIFSIGNALED(status))
			message(MESSAGE, "child process %d killed by signal %s", pid, signame(WTERMSIG(status)));
	}
	if (pid < 0 && errno != ECHILD)
		message(SYS|ERROR, "cannot waitpid");
}

void
actionOnSignal(int sig)
{
	signal(sig, SIG_IGN);
	getTime();
	message(MESSAGE, "receiving signal %s", signame(sig));

	switch(sig) {
	case SIGIO:
		// bad, very bad and very dirty quick-fix
		// of overflowing rt-signals queue
		doRestart = YES;
		break;

	case SIGHUP:
		doRestart = YES;
		break;
	case SIGWINCH:
        if (master) doReExec = YES;
		break;
	case SIGQUIT:
		doShutdown = YES;
		break;
	case SIGUSR1:
		doReload = YES;
		break;
	case SIGUSR2:
		Debug = Debug ? OFF : ON;
		message(MESSAGE, "turning debug to %s", Debug ? "ON" : "OFF");
		break;
	case SIGCHLD:
		reapChildren();
		break;
	case SIGPIPE:
		break;
	case SIGBUS:
	case SIGSEGV:
		if (currentRequest) logRequest(currentRequest);
		abort();

	case SIGINT:
	case SIGTERM:
	case SIGILL:
	case SIGTRAP:
	case SIGXCPU:
	case SIGXFSZ:
		if (++doQuit > 10) {
			message(ERROR, "quit signal was not handled %d times, immediately dying", doQuit - 1);
			exit(1);
		}
		break;
	}
	signal(sig, actionOnSignal);
}

int
sendSignal(pid_t pid, int s)
{
	int err;
	int result;
	uid_t uid = getuid();
	uid_t euid = geteuid();

	if (uid != euid && seteuid(uid)) message(SYS|ERROR, "cannot seteuid(%d)", uid);
	result = kill(pid, s);
	err = errno;
	if (uid != euid && seteuid(euid)) message(SYS|ERROR, "cannot seteuid(%d)", euid);
	errno = err;
	if (result) message(SYS|ERROR, "cannot send signal %s to %d", signame(s), pid);
	return result;
}


int
reExec(int argc, char** argv)
{
	int i;
	char* bin;
	char** a = MALLOC(sizeof(char*) * (argc + 1));

	unlink(fmt("%s/%s", PIDS, progName));
	message(MESSAGE, "reexecuting %s", progName);
	message(CLOSE_LOGS, NULL);

	// close pid-file
	for (i = 3; i < 10; i++) close(i);

	for (i = 0; i < argc; i++) a[i] = argv[i];
    a[i] = NULL;

	if ((bin = strrchr(a[0], '/'))) bin++;
	else bin = a[0];
	execv(fmt("bin/%s", bin), a);
	return message(SYS|ERROR, "cannot exec bin/%s", bin);
}


int
httpd(int argc, char** argv, char daemon)
{
	time(&start);
    if (wrapper) master = NO;

	if (!writePidFile(progName = "httpd", daemon, NO)) return FAILED;

	doRestart = doReload = doShutdown = doQuit = doReExec = NO;
	if (work(argc - 1, argv + 1)) {
#if DEBUGLOG
		showPtrs(YES);
#endif
		if (doReExec) return reExec(argc, argv);
		message(MESSAGE, "%d is shutting down", myPID);
        return OK;
	}

	return message(MESSAGE, "%d is exited due to some init error", myPID);
}

int
main(int argc, char** argv)
{
	int i;
	char daemon = YES;
	progName = "httpd";

	signal(SIGINT,   actionOnSignal);
	signal(SIGQUIT,  actionOnSignal);
	signal(SIGTERM,  actionOnSignal);
	signal(SIGSEGV,  actionOnSignal);
	signal(SIGHUP,   actionOnSignal);
	signal(SIGUSR1,  actionOnSignal);
	signal(SIGUSR2,  actionOnSignal);
	signal(SIGWINCH, actionOnSignal);
	signal(SIGILL,   actionOnSignal);
	signal(SIGTRAP,  actionOnSignal);
	signal(SIGBUS,   actionOnSignal);
	signal(SIGCHLD,  actionOnSignal);
	signal(SIGURG,   actionOnSignal);
	signal(SIGIO,    actionOnSignal);
	signal(SIGPIPE,  SIG_IGN);

	if (chdir(ROOT)) return message(SYS|ERROR, "cannot chdir to %s", ROOT);

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			int j;
			for (j = 0; argv[i][j]; j++) {
				switch (argv[i][j]) {
				case 'f': daemon = NO; break;
				case 'w': wrapper = YES; break;
				case 't':
					if (readConfig(argc - 1, argv + 1))
						return puts("Config OK") & 0;
					puts("Config file contains errors, see logs/current/error.log for details");
					return 2;
				case 'v':
					printf("Server version: %s, built: %s\n"
						   "Configured as: ./configure %s\n",
						   VERSION, built, CONFIGURE);
					return 1;
				case 'h':
					printf("Server version: %s, built: %s\n"
						   "Usage: %s [-options]\n"
						   "  -f	   don't daemonize (work in foreground)\n"
						   "  -w	   use wrapper\n"
						   "  -c NAME  use config file NAME\n"
						   "  -t	   test config file and exit\n"
						   "  -v	   show version and exit\n"
						   "  -h	   show this help page and exit\n",
						   VERSION, built, argv[0]);
					return 1;
				}
			}
		}
	}

	if (wrapper) {
		if (!writePidFile((progName = "httpd-wrapper"), daemon, NO)) return 1;

		while(TRUE) {
			int pid = fork();
			if (pid < 0) return !message(SYS|ERROR, "cannot fork");
			if (!pid) return !httpd(argc, argv, NO);
			message(MESSAGE, "launching httpd [%d]", pid);
			while (!childExits) {
				int s = 0;
				if (doQuit) s = SIGTERM;
				else if (doShutdown) s = SIGQUIT;
				else if (doReExec) s = SIGQUIT;
				else if (doRestart) s = SIGHUP;
				else if (doReload) s = SIGUSR1;

				doReload = doRestart = NO;
				if (s) sendSignal(pid, s);
				if (doShutdown || doReExec) break;
				pause();
			}
			if (childExits) {
				message(MESSAGE, "httpd exits");
                childExits = NO;
			}

			if (doQuit || doShutdown) return 0;
			if (doReExec) return !reExec(argc, argv);
			sleep(1);
		}
	}

	return !httpd(argc, argv, daemon);
}


