/* $Id: proxy.c,v 1.51 2008/09/20 23:04:02 maxim Exp $
 *
 */

#define _GNU_SOURCE

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>

#include "configure.h"

#if SUPPORT_PROXY
#include "common/settings.h"
#include "common/misc.h"
#include "common/str.h"

#include "core.h"
#include "httpd.h"
#include "config.h"
#include "flood.h"
#include "proxy.h"

#if SUPPORT_CACHE
#include "cache.h"
#endif

#if MODEL == MODEL_KEVENT
#include <sys/event.h>
#elif MODEL == MODEL_EPOLL
#include <sys/epoll.h>
#endif

#define REQUESTS_STEP 32


static char *states[] = {
	"UNDEF",
	"CONNECT",
	"WRITE_HEADER_BE",
	"WRITE_POST",
	"READ_RESPONSE",
	"READ_RESPONSE_H",
	"COPY_HEADER",
	"READ_BODY",
	"WAIT_MEMORY",
	"FINISHED",
	"KEEP_ALIVE_BE",
	"FAULT"
};


static ProxyRequest* requests = NULL;
static ProxyRequest* tail = NULL;
static ProxyRequest* freeRequests = NULL;
static ProxyRequest** requestsBlocks = NULL;
static int nRequests = 0, maxRequests = 0, nRequestsBlocks = 0, failed = 0;

#if SUPPORT_ELAPSED
#include "common/pairs.h"
static Pairs* elapsed;

typedef struct Elapsed_ {
	int count;
	int elapsed;
} Elapsed;


Elapsed*
freeElapsed(Elapsed* e)
{
	return FREE(e);
}
#endif

char
initProxyRequests()
{

	int n = 0;
	Backend* backend;
	for (n = 0, backend = config->backends; backend; n++, backend = backend->next);
	n = (n * config->proxyConnections / REQUESTS_STEP) + 1;

	if (!(requestsBlocks = REALLOC(requestsBlocks, n * sizeof(ProxyRequest*))))
		return FAILED;

#if SUPPORT_ELAPSED
	elapsed = newPairs(256, (void*)freeElapsed);
#endif
	return OK;
}

void
freeProxyRequests()
{
	int i;
	ProxyRequest* r = requests;
	while (r) {
		ProxyRequest* next = r->next;
		closeProxy(r);
		r = next;
	}

	for (i = 0; i < nRequestsBlocks; i++)
		FREE(requestsBlocks[i]);
	requestsBlocks = FREE(requestsBlocks);
	freeRequests = NULL;

	nRequests = maxRequests = nRequestsBlocks = failed = 0;

}

void
closeProxyAliveRequests()
{
	ProxyRequest* r = requests;
	debug("closeProxyAliveRequests()");
	while (r) {
		ProxyRequest* next = r->next;
		if (r->state == KEEP_ALIVE_BE) closeProxy(r);
		r = next;
	}
}

void
resetTodayProxy()
{
	maxRequests = failed = 0;
}

int
getProxyStats(char* buffer, int size, Request* r)
{
	int l = 0;
	int show = r->query && strstr(r->query, "proxy");

	l += SNPRINTF(buffer + l, size - l, "\n"
				  "<a href=\"%s%s\">Proxy requests</a> (active/max/failed): %s / %s / %s\n",
				  r->uri, show ? "" : "?proxy",
				  itos(nRequests), itos(maxRequests), itos(failed));
	if (show && size > l) {
		ProxyRequest* r;
		if (requests)
			l += SNPRINTF(buffer + l, size - l,
						  "%-15s %-15s %4s [%5s/%-5s] %s\n",
						  "State", "IP", "Time", "Recv", "Total", "URL");
		for (r = requests; r && size > l; r = r->next) {
			if (r->request) {
				l += SNPRINTF(buffer + l, size - l,
							  "%-15s %-15s %4s [%5s/%-5s] %s%s%s%s\n",
							  states[r->state],
							  ip2string(r->request->ip),
							  humanPeriod(now - r->start),
							  humanSize(r->request->received),
							  r->request->length < 0 ? "?" : humanSize(r->request->length),
							  r->request->host?:"-", r->request->uri,
							  r->request->query?"?":"", r->request->query?:"");
			} else if (r->state == KEEP_ALIVE_BE) {
				l += SNPRINTF(buffer + l, size - l,
							  "%-15s %-15s %4s [%5s/%-5s] %s:%d\n",
							  states[r->state],
							  inet_ntoa(r->proxy->backend->addr),
							  humanPeriod(now - r->start),
							  "?", "?",
							  r->proxy->backend->name, r->proxy->backend->port);
			} else {
				l += SNPRINTF(buffer + l, size - l, "%-15s %-15s %4s\n",
							  states[r->state], "?",
							  humanPeriod(now - r->start));
			}
		}
	}
	return l;
}

#if SUPPORT_LANGUAGES
char*
findLangInQuery(char* query)
{
	static char result[SHORT_STR];
	char* s = query;
	while ((s = strstr(s, "lang="))) {
		if (s == query || s[-1] == '&' || s[-1] == ';' || s[-1] == ' ') {
			char* e;
			char saved = 0;
			s += strlen("lang=");
			e = strpbrk(s, "&; ");
			if (e) {
				saved = *e;
				*e = 0;
			}
			if (getPair(config->languages, s)) {
				strcpy(result, s);
				if (e) *e = saved;
				return result;
			}
			break;
		}
		s += strlen("lang=");
	}
	return NULL;
}

char*
getPreferredLanguage(Request* r)
{
	static char accepts[MEDIUM_STR];
	char* lang;

	if (r->query && (lang = findLangInQuery(r->query))) return lang;
	if (r->cookie && (lang = findLangInQuery(r->cookie))) return lang;

	if (!(r->acceptLanguage && *(r->acceptLanguage))) return NULL;
	strncpy_(accepts, r->acceptLanguage, MEDIUM_STR);
	strtolower(accepts);

	for (lang = strtok(accepts, ","); lang; lang = strtok(NULL, ",")) {
		char* e = strchr(lang, ';');
		if (e) *e = 0;
		lang = trimSpaces(lang);
		if (getPair(config->languages, lang)) return lang;
		if ((e = strchr(lang, '-')) || (e = strchr(lang, '_'))) {
			*e = 0;
			if (getPair(config->languages, lang)) return lang;
		}
	}
	return NULL;
}

#endif

int
fillProxyHeaders(ProxyRequest* pr, char* buffer, int size)
{
	int l = 0;
	char* b = buffer;
	char location[MAX_URI];
	Request* r = pr->request;

	size -= 2;

	debug("%d> fillProxyHeaders()", pr->fd);

	l = catstr(location, MAX_URI, pr->proxy->location, r->alias->exactLocation ? "" : escape(r->path), NULL);
	if (!l) return 0;

#if SUPPORT_LANGUAGES
	if (r->alias->substLanguage == 2) {
		char* lang = getPreferredLanguage(r);

		if (lang || (lang = config->defaultLanguage)) {
			l += catstr(location + l, MAX_URI - l,
						location[l - 1] == '/' ? "index.html." : ".", lang, NULL);
			r->acceptLanguage = addStringToBuffer(r, lang, strlen(lang) + 1);
		}
	}
#endif

	if (r->params)
		l += catstr(location + l, MAX_URI - l, ";", r->params, NULL);
	if (r->query)
		l += catstr(location + l, MAX_URI - l, "?", r->query, NULL);

	l = catstr(b, size, methods[r->method], " ", location, " HTTP/1.0" RN,
			   "X-Real-IP: ", ip2string(r->ip), RN, NULL);
	if (!l) return 0;

	if (pr->proxy->host)
		l += catstr(b + l, size - l, "Host: ", pr->proxy->host, RN, NULL);
	if (r->host)
		l += catstr(b + l, size - l, pr->proxy->host ? "X-" : "", "Host: ", r->host, RN, NULL);
	if (r->userAgent)
		l += catstr(b + l, size - l, "User-Agent: ", r->userAgent, RN, NULL);
	if (r->referrer)
		l += catstr(b + l, size - l, "Referer: ", r->referrer, RN, NULL);
	if (r->authorization)
		l += catstr(b + l, size - l, "Authorization: ", r->authorization, RN, NULL);
#if SUPPORT_LANGUAGES
	if (r->alias->substLanguage == OK) {
		char* lang = getPreferredLanguage(r);
		if (lang) {
			if (!r->acceptLanguage || strcmp(r->acceptLanguage, lang))
				r->acceptLanguage = addStringToBuffer(r, lang, strlen(lang) + 1);
		} else {
			r->acceptLanguage = config->defaultLanguage;
		}
		l += catstr(b + l, size - l, "Accept-Language: ", r->acceptLanguage, RN, NULL);
	} else if (r->acceptLanguage) {
		l += catstr(b + l, size - l, "Accept-Language: ", r->acceptLanguage, RN, NULL);
	}

#endif
	if (r->cookie)
		l += catstr(b + l, size - l, "Cookie: ", r->cookie, RN, NULL);
	if (r->inContentType)
		l += catstr(b + l, size - l, "Content-Type: ", r->inContentType, RN, NULL);
	if (r->inContentLength >= 0)
		l += catstr(b + l, size - l, "Content-Length: ", itoa(r->inContentLength), RN, NULL);
	if (r->rangeFrom >= 0 || r->rangeTo >= 0) {
		if (r->rangeFrom >= 0) {
			if (r->rangeTo >= 0)
				l += catstr(b + l, size - l, "Range: bytes=", itoa(r->rangeFrom), "-", itoa(r->rangeTo), RN, NULL);
			else
				l += catstr(b + l, size - l, "Range: bytes=", itoa(r->rangeFrom), "-", RN, NULL);
		} else {
			l += catstr(b + l, size - l, "Range: bytes=-", itoa(r->rangeTo), RN, NULL);
		}
	}
	if (r->ims > 0)
		l += catstr(b + l, size - l, "If-Modified-Since: ", time2str(r->ims, HTTP_TIME), RN, NULL);
	if (r->iums > 0)
		l += catstr(b + l, size - l, "If-Unmodified-Since: ", time2str(r->iums, HTTP_TIME), RN, NULL);
	if (r->ifRange > 0)
		l += catstr(b + l, size - l, "If-Range: ", time2str(r->ifRange, HTTP_TIME), RN, NULL);
	if (r->forwardedFor)
		l += catstr(b + l, size - l, "X-Forwarded-For: ", ip2string(r->forwardedFor), RN, NULL);
#if SUPPORT_COUNTRY
	l += catstr(b + l, size - l, "X-Country: ",
				id2code(countries, r->country < 0 ? (r->country = ip2id(countries, r->ip)) : r->country),
				RN, NULL);
	l += catstr(b + l, size - l, "X-Region: ",
				id2code(countries, r->country), "/",
				(r->country == ruID) ? id2code(regions, ip2id(regions, r->ip)) : "",
				RN, NULL);
#endif
#if MAX_OTHER_HEADERS
	{
		int i;
		for (i = 0; i < r->inHeadersCount; i++) {
			int n = strlen(r->inHeaders[i]);
			if (l + n + 2 < size) {
				memcpy(b + l, r->inHeaders[i], n); l += n;
				memcpy(b + l, RN, 2); l += 2;
			}
		}
	}
#endif
	if (r->alias->proxyKeepalive)
		l += catstr(b + l, size - l, "Connection: Keep-Alive", RN, NULL);

	memcpy(b + l, RN, 2);
	return l + 2;
}


int
handleProxyRequest(Request* request)
{
	int fd;
	Proxy* pr = request->alias->proxy;
	Proxy* startProxy = pr;
	ProxyRequest* r;
	int allChecked = NO;
	int connected = NO;
	struct sockaddr_in address;

#if SUPPORT_ANTIFLOOD
	if (!request->trusted && request->alias->antiflood &&
		!addIPRequest(request->ip)) return 503;
#endif

	if (!pr) {
		message(ERROR, "no backend defined for %s%s", request->host?:"-", request->uri?:"-");
		failed++;
		return 502;
	}

backendsLoop:
	do {
		if (pr->backend->state) break;
		if (pr->backend->last + config->timeout < now) break;
		pr = pr->next;
		allChecked = YES;
	} while (pr != startProxy);

	allChecked = allChecked && (pr == startProxy || pr->next == startProxy);

	if (!(*(pr->location) || *(request->path))) {
		char location[MAX_URI];
		int n = catstr(location, MAX_URI, "http://",
					   request->host?:request->server->name, request->uri,
					   "/", NULL);
		if (!n) return 414;
		if (request->params)
			n += catstr(location + n, MAX_URI - n, ";", request->params, NULL);
		if (request->query)
			n += catstr(location + n, MAX_URI - n, "?", request->query, NULL);
		if (!(request->location = addStringToBuffer(request, location, n + 1)))
			return 414;
		return 301;
	}

	pr->backend->last = now;

	if ((r = pr->backend->alive)) {
		debug("%d> using keeped alive", r->fd);

		r->request = request;
		r->proxy = pr;

		request->state = HANDLING;
		request->proxy = r;

		request->alias->proxy = pr->next;
		request->contentType = NULL;
		request->contentEnc = NULL;
		pr->backend->requests++;

		if (r->nextAlive) r->nextAlive->prevAlive = NULL;
		pr->backend->alive = r->nextAlive;
		r->nextAlive = NULL;


		r->last = now + config->timeout;
		if (requests != r) {
			if (r->next) r->next->prev = r->prev;
			else tail = r->prev;
			r->prev->next = r->next;
			r->next = requests;
			r->prev = NULL;
			requests->prev = r;
			requests = r;
		}

		r->state = WRITE_HEADER_BE;
		proxy(r);
		if (!r->fd) {
			debug("%d> error while initiating proxy", r->fd);
			return 503;
		}
		return 0;
	}

	if (pr->backend->nRequests == config->proxyConnections) {
		message(ERROR, "connection to %s:%d failed: too many requests", inet_ntoa(pr->backend->addr), pr->backend->port);
		goto backendError;
	}

	if ((fd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		message(SYS|ERROR, "cannot create socket");
		failed++;
		return 503;
	}
	debug("%d> handleProxyRequest(%s %s%s): %d", request->fd, ip2string(request->ip), request->host?:"-", request->uri?:"-", fd);

	if (!fd || fd >= fdLimit) {
		if (close(fd)) message(SYS|ERROR, "cannot close proxy fd");
		message(ERROR, "connection to %s:%d failed: bad fd (%d)", inet_ntoa(pr->backend->addr), pr->backend->port, fd);
		failed++;
		return 503;
	}

	//	fcntl(fd, F_SETFD, FD_CLOEXEC);
#if MODEL == MODEL_RTSIG
	fcntl(fd, F_SETSIG, rtSignal);
	fcntl(fd, F_SETOWN, myPID);
	fcntl(fd, F_SETFL, O_NONBLOCK|O_ASYNC);
#else
	fcntl(fd, F_SETFL, O_NONBLOCK);
#endif

	memset(&(address), 0, sizeof(address));
	address.sin_addr = pr->backend->addr;
	address.sin_port = htons(pr->backend->port);
	address.sin_family = AF_INET;

	if (connect(fd, (struct sockaddr*) &address, sizeof(address)) < 0) {
		if (errno != EINPROGRESS) {
			message(SYS|ERROR, "cannot connect to %s:%d", inet_ntoa(pr->backend->addr), pr->backend->port);
			if (close(fd)) message(SYS|ERROR, "cannot close proxy fd");
			goto backendError;
		}
	} else {
		connected = YES;
	}

	if (!(r = freeRequests)) {
		debug("handleProxyRequest: allocating new proxyRequests block");
		freeRequests = CALLOC(REQUESTS_STEP * sizeof(ProxyRequest));
		if (!freeRequests) {
			message(ERROR, "connection to %s:%d failed: not enough memory", inet_ntoa(pr->backend->addr), pr->backend->port);
			failed++;
			return 503;
		}
		for (r = freeRequests; r < freeRequests + REQUESTS_STEP - 1; r++)
			r->next = r + 1;
		r = requestsBlocks[nRequestsBlocks++] = freeRequests;
	}

	freeRequests = freeRequests->next;
	if ((r->next = requests)) r->next->prev = r;
	else tail = r;
	requests = r;
	r->prev = NULL;

	r->type = PROXY_REQUEST;
	r->state = CONNECT;
	r->fd = fd;
	r->length = -1;
	r->last = now + config->timeout;
	r->start = now;
	r->rangeFrom = r->rangeTo = -1;
	r->request = request;
	r->proxy = pr;

	if (fd > maxFD) maxFD = fd;
	if (++nRequests > maxRequests) maxRequests = nRequests;
	if (++(pr->backend->nRequests) > pr->backend->maxRequests) pr->backend->maxRequests = pr->backend->nRequests;
	request->state = HANDLING;
	request->proxy = r;
	fds[fd].proxy = r;


	request->alias->proxy = pr->next;
	request->contentType = NULL;
	request->contentEnc = NULL;
	pr->backend->requests++;

	if (connected) {
		debug("%d> immediate connect", fd);
		proxy(r);
		if (!r->fd) {
            debug("%d> error while initiating proxy", fd);
			return 503;
		}
	} else {
#if MODEL == MODEL_POLL
		r->watchWrite = YES;
#elif MODEL == MODEL_KEVENT
		r->watchWrite = YES;
		watchEvent(fd, EVFILT_WRITE, EV_ADD, 0, 0);
		//strange modern FreeBSD behaivor, don't use oneshot
		//watchEvent(fd, EVFILT_WRITE, EV_ADD | EV_ENABLE | EV_ONESHOT, 0, 0);
#elif MODEL == MODEL_EPOLL
		r->watchWrite = YES;
		watchEvent(fd, EPOLLOUT /*|EPOLLONESHOT*/, EPOLL_CTL_ADD);
#endif
	}

	return 0;

backendError:
	if (pr->backend->state) {
		message(MESSAGE, "backend %s:%d is DOWN", inet_ntoa(pr->backend->addr), pr->backend->port);
		pr->backend->state = FAILED;
	}

	if (!allChecked) {
		pr = pr->next;
		goto backendsLoop;
	}
	failed++;
	request->alias->proxy = pr->next;

#if SUPPORT_CACHE
	{
		int result;
		if (request->alias->cache && request->alias->cache->onFault &&
			(result = handleCached(request)) != 404) {
			debug("%d> handleProxyRequest(%s %s%s): result from OnFault-cache",
				  request->fd, ip2string(request->ip), request->host?:"-", request->uri?:"-");
			return result;
		}
	}
#endif

	return 503;
}

#define COPY_FIELD(name) if (r->name >= start && r->name < end) r->name += diff

int
increaseResponseBuffer(ProxyRequest* r)
{
	off_t diff;
	Buffer* src = r->buffer;
	char* start = src->buffer;
	char* end = src->buffer + src->count;
	if (!(r->buffer = newBuffer(src->size << 1))) {
		r->buffer = src;
		return FAILED;
	}
	memcpy(r->buffer->buffer, src->buffer,
		   r->buffer->count = src->count);
	r->buffer->start = src->start;

	diff = r->buffer->buffer - src->buffer;
	COPY_FIELD(contentType);
	COPY_FIELD(contentEnc);
	COPY_FIELD(location);
	COPY_FIELD(auth);
	COPY_FIELD(setCookie);
#if MAX_OTHER_HEADERS
	{
		int i;
		for (i = 0; i < r->outHeadersCount; i++)
			COPY_FIELD(outHeaders[i]);
	}
#endif
	freeBuffer(src);
	return OK;
}


#define COPY_TO_REQUEST(name) if(r->name) request->name = addStringToBuffer(request, r->name, strlen(r->name) + 1)

void
copyHeader(ProxyRequest* r)
{
	Request* request = r->request;
	request->modified = r->modified;
	request->expires = r->expires;
	request->length = r->length;
	request->acceptRanges = r->acceptRanges;
	request->status = r->status;
	request->rangeFrom = r->rangeFrom;
	request->rangeTo = r->rangeTo;
	request->rangeTotal = r->rangeTotal;
	COPY_TO_REQUEST(contentType);
	COPY_TO_REQUEST(contentEnc);
	COPY_TO_REQUEST(location);
	COPY_TO_REQUEST(auth);
	COPY_TO_REQUEST(setCookie);
#if MAX_OTHER_HEADERS
	{
		int i;
		request->outHeadersCount = r->outHeadersCount;
		for (i = 0; i < r->outHeadersCount; i++)
			COPY_TO_REQUEST(outHeaders[i]);
	}
#endif
}

char*
getResponseLine(ProxyRequest* r)
{
	if (!r || r->fd <= 0) {
		message(ERROR, "getLine for bad request: %s", r ? "not connected" : "NULL");
		return NULL;
	}

	while (1) {
		int cnt;
		if (r->buffer->count - r->buffer->start > 0) {
			char* line = r->buffer->buffer + r->buffer->start;
			char* eol;
			if ((eol = memchr(line, '\n', r->buffer->count - r->buffer->start))) {
				r->buffer->start = eol + 1 - r->buffer->buffer;
				*eol-- = 0;
				while (eol >= line && isSpace(*eol)) *eol-- = 0;
				return line;
			}
		}

		if (r->buffer->count == r->buffer->size)  {
			if (r->buffer->size < MAXIMUM_BUFFER_SIZE) {
				if (!increaseResponseBuffer(r)) return NULL;
			} else {
				message(ERROR, "proxy buffer overflow while reading from %s:%d", inet_ntoa(r->proxy->backend->addr), r->proxy->backend->port);
				return NULL;
			}
		}

		errno = 0;
		cnt = read(r->fd, r->buffer->buffer + r->buffer->count, r->buffer->size - r->buffer->count);
		debug("%d> reading %d bytes (%s)", r->fd, cnt, errno?strerror(errno):"OK");
		if (cnt <= 0) {
			if (errno == EWOULDBLOCK) return NULL;
			if (cnt)
				message(SYS|ERROR, "cannot read from %s:%d", inet_ntoa(r->proxy->backend->addr), r->proxy->backend->port);
			else
				message(ERROR, "cannot read from %s:%d: %s", inet_ntoa(r->proxy->backend->addr), r->proxy->backend->port,
						"connection is closed");
			return NULL;
		}
		r->buffer->count += cnt;
	}
}


int
parseResponseLine(char* line, ProxyRequest* r)
{
	char* c;

	debug("parseResponseLine(%s)", line);

	if (strncmp(line, "HTTP/", 5)) return 502;

	c = line + 5;
	if (*c != '1' && *c != '0') return 502;
	r->major = 1;
	c++;

	if (*c++ != '.') return 502;
	if (!isDigit(*c)) return 502;
	r->minor = *c++ - '0';

	if (*c++ != ' ') return 502;

	r->status = atol(c);
	if (r->status < 1) r->status = 502;

	r->keepalive = r->minor;

	return 0;
}

int
parseResponseHeaderLine(char* line, ProxyRequest* r)
{
	char* value = strchr(line, ':');
	if (!value) return 502;
	*value++ = 0;
	while (isSpace(*value)) value++;
	if (!*value) return 0;

	strtolower(line);

	if (!strcmp(line, "content-type")) {
		r->contentType = strtolower(value);
	} else if (!strcmp(line, "content-encoding")) {
		if (strcmp(value, "identity"))
			r->contentEnc = strtolower(value);
	} else if (!strcmp(line, "content-length")) {
		r->length = atol(value);
	} else if (!strcmp(line, "accept-ranges")) {
		r->acceptRanges = !strcmp(value, "bytes");
	} else if (!strcmp(line, "content-range")) {
		ssize_t from, to, total;
		char* end;
		if (strncmp(value, "bytes ", strlen("bytes "))) return 0;
		value += strlen("bytes ");
		errno = 0;
		from = strtol(value, &end, 10);
		if (end == value || *end++ != '-' || errno) return 0;
		to = strtol(value = end, &end, 10);
		if (end == value || *end++ != '/' || errno) return 0;
		total = strtol(value = end, &end, 10);
		if (end == value || *end != 0 || errno) return 0;

		r->rangeFrom = from;
		r->rangeTo = to;
		r->rangeTotal = total;
	} else if (!strcmp(line, "last-modified")) {
		r->modified = timerfc(strtolower(value));
		if (r->modified == (time_t) -1) r->modified = 0;
	} else if (!strcmp(line, "expires")) {
		r->expires = timerfc(strtolower(value));
		if (r->expires < 1) r->expires = EXPIRES_PAST;
	} else if (!strcmp(line, "location")) {
		if (r->request->host && r->proxy) {
			char* withPort = fmt("http://%s:%d/", r->request->host, r->proxy->backend->port);
			int n = strlen(withPort);
			if (!strncmp(value, withPort, n))
				strcpy(value + 7 + strlen(r->request->host), value + n - 1);
		}
		r->location = value;
	} else if (!r->setCookie && !strcmp(line, "set-cookie")) {
		r->setCookie = value;
	} else if (!strcmp(line, "www-authenticate")) {
		r->auth = value;
	} else if (!strcmp(line, "connection")) {
		strtolower(value);
		if (r->minor) {
			if (!strcmp(value, "close"))
				r->keepalive = NO;
		} else {
			if (!strcasecmp(value, "keep-alive"))
				r->keepalive = YES;
		}
	} else if (!strcmp(line, "server")) {
		//ignored;
	} else if (!strcmp(line, "date")) {
		//ignored;
	} else if (!strcmp(line, "keep-alive")) {
		//ignored;
	} else {
#if MAX_OTHER_HEADERS
		if (r->outHeadersCount < MAX_OTHER_HEADERS) {
			line[strlen(line)] = ':';
			r->outHeaders[r->outHeadersCount++] = line;
		}
#endif
	}
	return 0;
}


void
cleanupProxy()
{
	while(tail && now > tail->last) {
		debug("%d> timeout in %s", tail->fd, states[tail->state]);
		if (tail->request) {
			message(MESSAGE, "proxy timeout in %s for %s (%s%s%s%s)",
					states[tail->state], ip2string(tail->request->ip),
					tail->request->host?:"?", tail->request->uri,
					tail->request->query?"?":"", tail->request->query?:"");
		} else if (tail->state == KEEP_ALIVE_BE) {
			closeProxy(tail);
			continue;
		} else {
			message(MESSAGE, "proxy timeout in %s", states[tail->state]);
		}

		if (tail->state == CONNECT && tail->proxy->backend->state) {
			tail->proxy->backend->state = FAILED;
			tail->proxy->backend->last = now;
			message(MESSAGE, "backend %s:%d is DOWN",
					inet_ntoa(tail->proxy->backend->addr),
					tail->proxy->backend->port);
		}

		tail->status = 504;
		tail->state = FAULT;
		proxy(tail);
	}

#if SUPPORT_ELAPSED
	if (elapsed->count) {
		int fd = open("data/httpd/elapsed", O_RDWR|O_CREAT, 0644);
		if (fd < 0) {
			message(SYS|ERROR, "cannot open %s", "data/httpd/elapsed");
		} else {
			char buf[LARGE_TEXT];
			int n;
			Pair* p;
			flock(fd, LOCK_EX);
			if ((n = read(fd, buf, LARGE_TEXT - 1)) > 0) {
				char* end;
				char* s = buf;
				buf[n] = 0;
				do {
					char name[MEDIUM_STR];
					char count[MEDIUM_STR];
					char total[MEDIUM_STR];
					Elapsed *e;
					if ((end = strchr(s, '\n'))) *end++ = 0;
					split(s, " ", MEDIUM_STR, name, count, total, NULL);
					if ((e = getPair(elapsed, name))) {
						e->count += atoi(count);
						e->elapsed += atoi(total);
					}
				} while ((s = end));
				lseek(fd, 0, SEEK_SET);
			}
			if (elapsed->count > 1000) {
				ftruncate(fd, 0);
				freePairs(elapsed);
				elapsed = newPairs(256, (void*)freeElapsed);
			}
			n = 0;
			startListPairs(elapsed);
			while ((p = nextPair(elapsed))) {
				Elapsed* e = p->value;
				n += catstr(buf + n, LARGE_TEXT - n,
							p->name, " ", itoa(e->count), " ", itoa(e->elapsed), "\n", NULL);
				e->count = e->elapsed = 0;
			}
			write(fd, buf, n);
			close(fd);
		}
	}
#endif
}

ProxyRequest*
flushResponse(Request* r)
{
	debug("%d> flushResponse()", r->proxy->fd);
	if (currentRequest) return r->proxy;

	switch (r->state) {
	case HANDLING:
#if SUPPORT_CACHE
		if (r->alias->cache && r->status == 200 &&
			(r->expires == 0 || r->expires > now || r->alias->cache->onFault) &&
			r->length >= 0 && r->length <= (ssize_t)r->alias->cache->maxLength) {
			r->cached = newCacheEntry(r);
		}
#endif
		r->state = PREPARE_HEADER;
		http(r);
		break;

	case WAIT_DATA:
		http(r);
		break;

	case WRITE_HEADER:
		http(r);
		break;

#if SUPPORT_THROTTLE
	case THROTTLED:
#endif
	case WRITE_CONTENT:
		break;

	default:
		message(MESSAGE, "strange state of request: %d", r->state);
	}
	currentRequest = NULL;
	return r->proxy;
}


ProxyRequest*
closeProxy(ProxyRequest* r)
{
	if (!r || !r->state) return NULL;
	debug("%d> closeProxy()", r->fd);

#if SUPPORT_ELAPSED
	if (r->request && r->request->alias->elapsed &&
		r->request->path && r->exactStart.tv_sec) {
		struct timeval finish;
		Elapsed* e;
		int l = strlen(r->request->path);
		char* key = l > 4 && strcmp(r->request->path + l - 4, ".php") == 0 ? "php" : "html";
		gettimeofday(&finish, NULL);
		l = ((finish.tv_sec - r->exactStart.tv_sec) * 1000000 + (finish.tv_usec - r->exactStart.tv_usec))/1000;
		if (l < 0) {
			message(MESSAGE, "strange: negative elapsed: %d.%d > %d.%d",
					r->exactStart.tv_sec, r->exactStart.tv_usec,
					finish.tv_sec, finish.tv_usec);
			l = 0;
		}
		if (!(e = getPair(elapsed, key))) {
			e = CALLOC(sizeof(Elapsed));
			setPair(elapsed, key, e);
		}
		e->count++;
		e->elapsed += l;
	}
#endif

	if (r->request) r->request->proxy = NULL;
	freeBuffer(r->buffer);

	if (r->keepalive && r->state == FINISHED) {
		int fd = r->fd;
		Proxy* pr = r->proxy;
		ProxyRequest* next = r->next;
		ProxyRequest* prev = r->prev;
#if MODEL != MODEL_RTSIG
		char wRead = r->watchRead, wWrite = r->watchWrite;
#endif

		debug("%d> keeping alive", r->fd);

		memset(r, 0, sizeof(ProxyRequest));

		r->type = PROXY_REQUEST;
		r->state = KEEP_ALIVE_BE;
		r->fd = fd;
		r->length = -1;
		r->start = now;
		r->rangeFrom = r->rangeTo = -1;
		r->proxy = pr;
		r->next = next;
		r->prev = prev;

#if MODEL != MODEL_RTSIG
		r->watchRead = wRead;
		r->watchWrite = wWrite;
#endif

		if ((r->nextAlive = pr->backend->alive)) r->nextAlive->prevAlive = r;
		pr->backend->alive = r;

		return r;
	}

	if (r->fd > 0) {
#if MODEL == MODEL_KEVENT
        if (r->watchRead || r->watchWrite) unwatchEvent(r->fd);
#endif
		fds[r->fd].proxy = NULL;
		if (close(r->fd) < 0)
			message(SYS|ERROR, "cannot close proxy fd %d", r->fd);
	}

	if (r->next) r->next->prev = r->prev;
	else tail = r->prev;
	if (r->prev) r->prev->next = r->next;
	else requests = r->next;
	r->proxy->backend->nRequests--;


	if (r->state == KEEP_ALIVE_BE) {
		debug("%d> removing from alive list", r->fd);
		if (r->nextAlive) r->nextAlive->prevAlive = r->prevAlive;
		if (r->prevAlive) r->prevAlive->nextAlive = r->nextAlive;
		else r->proxy->backend->alive = r->nextAlive;
	}

	memset(r, 0, sizeof(ProxyRequest));

	r->next = freeRequests;
	freeRequests = r;
	nRequests--;
	return NULL;
}

void
proxy(ProxyRequest* r)
{
	Request* request = r->request;
	char* line;
	char watchRead = 0;
	char watchWrite = 0;

	while(1) {
		switch (r->state) {
		case UNDEF:
			debug("%d> proxy: %s", r->fd, states[r->state]);
			message(ERROR, "strange: proxy in UNDEF state (%d)", r->fd);
			goto end;

		case CONNECT: {
			int err;
			socklen_t l;
			debug("%d> proxy: %s", r->fd, states[r->state]);

			err = 0;
			l = sizeof(err);
			if (getsockopt(r->fd, SOL_SOCKET, SO_ERROR, &err, &l) < 0 || err) {
				if (err) errno = err;
				message(SYS|ERROR, "cannot connect to %s:%d", inet_ntoa(r->proxy->backend->addr), r->proxy->backend->port);
				r->state = FAULT;
				if (r->proxy->backend->state) {
					message(MESSAGE, "backend %s:%d is DOWN", inet_ntoa(r->proxy->backend->addr), r->proxy->backend->port);
					r->proxy->backend->state = FAILED;
				}
				break;
			}
			r->state = WRITE_HEADER_BE;
			//break;
		}

		case WRITE_HEADER_BE: {
			ssize_t l;
			char header[LARGE_TEXT];
			ssize_t count;
			char writePost = NO;

			debug("%d> proxy: %s", r->fd, states[r->state]);
			l = fillProxyHeaders(r, header, LARGE_TEXT);
			if (!l) {
				message(ERROR, "too long proxy request: %s", r->request->uri);
				r->status = 414;
				r->state = FAULT;
				break;
			}

			if (request->inContentLength > 0) {
				if (!request->postBuffer) {
					message(ERROR, "wow! post without postbuffer");
				} else if (request->inContentLength <= LARGE_TEXT - l) {
					memcpy(header + l, request->postBuffer->buffer, request->inContentLength);
					l += request->inContentLength;
				} else {
					writePost = YES;
				}
			}

			l -= r->offset;

#if SUPPORT_ELAPSED
			if (request->alias->elapsed) gettimeofday(&r->exactStart, NULL);
#endif
			watchWrite = 1;
			if ((count = write(r->fd, header + r->offset, l)) <= 0) {
				if (errno == EWOULDBLOCK) goto end;
				if (errno == ENOTCONN) {
					debug("do retry. cannot write to backend %s:%d", inet_ntoa(r->proxy->backend->addr), r->proxy->backend->port);
					goto end;
				}
				message(SYS|ERROR, "cannot write to backend %s:%d", inet_ntoa(r->proxy->backend->addr), r->proxy->backend->port);
				r->state = FAULT;
				break;
			}
			if (count < l) {
				debug("%d> incomplete header write to backend (%d of %d bytes)", r->fd, count, l);
				r->offset += count;
				goto end;
			}

			if (writePost) {
				r->state = WRITE_POST;
				break;
			}

			r->state = READ_RESPONSE;
			watchWrite = -1;
			watchRead = 1;
			goto end;
		}

		case WRITE_POST: {
			ssize_t cnt;
			cnt = write(r->fd, request->postBuffer->buffer + r->posted, request->inContentLength - r->posted);
			debug("%d> writing %d bytes (%s)", r->fd, cnt, errno?strerror(errno):"OK");
			if (cnt <= 0) {
				if (errno == EWOULDBLOCK) goto end;
				if (cnt < 0) message(SYS|ERROR, "cannot write to backend %s:%d", inet_ntoa(r->proxy->backend->addr), r->proxy->backend->port);
				r->state = FAULT;
				break;
			}
			r->posted += cnt;

			if (r->posted == request->inContentLength) {
				r->state = READ_RESPONSE;
				watchWrite = -1;
				watchRead = 1;
				goto end;
			}
			break;
		}

		case READ_RESPONSE:
			debug("%d> proxy: %s", r->fd, states[r->state]);

			if (!r->buffer) r->buffer = newBuffer(INITIAL_BUFFER_SIZE);

			if (!(line = getResponseLine(r))) {
				if (errno == EWOULDBLOCK) goto end;
				message(ERROR, "strange: cannot get response line");
				r->state = FAULT;
				break;
			}
			if (parseResponseLine(line, r)) {
				message(ERROR, "strange: cannot parse response line %s", line);
				r->status = 502;
				r->state = FAULT;
				break;
			}
			r->state = READ_RESPONSE_H;
			//break;

		case READ_RESPONSE_H:
			debug("%d> proxy: %s", r->fd, states[r->state]);
			while (1) {
				if (!(line = getResponseLine(r))) {
					if (errno == EWOULDBLOCK) goto end;
					message(ERROR, "strange: cannot get response header");
					r->state = FAULT;
					break;
				}
				if (!*line) {
					r->state = COPY_HEADER;
					break;
				}

				if (parseResponseHeaderLine(line, r)) {
					message(ERROR, "strange: cannot parse response header %s", line);
					r->status = 502;
					r->state = FAULT;
					break;
				}
			}
			break;

		case COPY_HEADER: {
			debug("%d> proxy: %s", r->fd, states[r->state]);
			copyHeader(r);

			if (!r->proxy->backend->state &&
				r->proxy->backend->nRequests <= config->proxyConnections / 2) {
				message(MESSAGE, "backend %s:%d is UP", inet_ntoa(r->proxy->backend->addr), r->proxy->backend->port);
				r->proxy->backend->state = OK;
			}

			if (r->length == 0 || r->status == 304 || request->method == HEAD) {
				r->state = FINISHED;
				break;
			}

			request->received = r->buffer->count - r->buffer->start;
			if (r->length > 0 && request->received >= r->length) {
				request->received = r->length;
				r->state = FINISHED;
			} else {
				r->state = READ_BODY;
			}

			request->contentSize = r->length < 0 ? BLOCK_SIZE : r->length < config->maxBuffer ? r->length : config->maxBuffer;
			if (request->contentSize < request->received)
				request->contentSize = request->received;
			request->content = CALLOC(request->contentSize);
			if (request->received) memcpy(request->content, r->buffer->buffer + r->buffer->start, request->received);
			r->buffer = freeBuffer(r->buffer);
			break;
		}

		case READ_BODY:
			debug("%d> proxy: %s", r->fd, states[r->state]);
			while (1) {
				ssize_t count = request->contentSize + request->offset - request->received;
				debug("READ_BODY: size: %d, off: %d, r: %d, s: %d", request->contentSize, request->offset, request->received, request->sent);
				if (count == 0) {
					if (request->contentSize + BLOCK_SIZE <= config->maxBuffer) {
						request->content = REALLOC(request->content, request->contentSize += BLOCK_SIZE);
						count += BLOCK_SIZE;
					} else {
						if (flushResponse(request) != r) {
							if (!r->state) return;
							message(MESSAGE, "strange: flushResponse != r and r->state in READ_BODY");
							r->state = FAULT;
							break;
						}
						r->state = WAIT_MEMORY;
						watchRead = -1;
						break;
					}
				}

				if ((count = read(r->fd, request->content + request->received - request->offset, count)) < 0) {
					debug("%d> cannot read: %s", r->fd, errno?strerror(errno):"OK");
					if (errno == EWOULDBLOCK) {
						if (now - r->start > 1 && flushResponse(request) != r) {
							if (!r->state) return;
							message(MESSAGE, "strange: flushResponse != r and r->state in READ_BODY");
							r->state = FAULT;
							break;
						}
						goto end;
					} else {
						message(SYS|ERROR, "cannot read from %s:%d", inet_ntoa(r->proxy->backend->addr), r->proxy->backend->port);
						r->state = request->received ? FINISHED : FAULT;
					}
					break;
				}
				debug("%d> reading %d bytes", r->fd, count);


				request->received += count;
				if (count == 0 || request->received == request->length) {
					if (request->method != HEAD) request->length = request->received;
					r->state = FINISHED;
					break;
				}
			}
			break;

		case WAIT_MEMORY:
			debug("%d> proxy: %s", r->fd, states[r->state]);
			if (request->contentSize + request->offset - request->received <= 0) goto end;
			r->state = READ_BODY;
			watchRead = 1;
			break;


		case FINISHED:
			debug("%d> proxy: %s", r->fd, states[r->state]);
			r = closeProxy(r);
			flushResponse(request);
			if (r) {
				watchRead = 1;
				watchWrite = -1;
				goto end;
			} else {
				return;
			}

		case KEEP_ALIVE_BE: {
			debug("%d> proxy: %s", r->fd, states[r->state]);
			closeProxy(r);
			return;
		}


		case FAULT:
			debug("%d> proxy: %s", r->fd, states[r->state]);
			if ((request = r->request) && request->proxy) {
				request->status = r->status >= 400 ? r->status : 503;
				if (request->contentSize) {
					request->content = FREE(request->content);
					request->contentSize = 0;
					request->state = CLOSE;
				} else {
#if SUPPORT_CACHE
					int result;
					if (request->alias->cache && request->alias->cache->onFault &&
						request->state == HANDLING &&
						(result = handleCached(request)) != 404) {
						debug("%d> proxy(%s %s%s): result from OnFault-cache",
							  request->fd, ip2string(request->ip), request->host?:"-", request->uri?:"-");
						request->status = result;
						request->state = PREPARE_HEADER;
					} else {
						request->state = request->state == HANDLING ? NOT_OK : CLOSE;
					}
#else
					request->state = request->state == HANDLING ? NOT_OK : CLOSE;
#endif
				}
				if (!currentRequest) {
					http(request);
					currentRequest = NULL;
				} else {
					r = closeProxy(r);
				}

			} else {
				r = closeProxy(r);
			}
			failed++;
			return;
		}
	}

end:

	if (r && r->state) {
#if MODEL == MODEL_KEVENT
		if (watchRead > 0) {
			if (!r->watchRead) {
				watchEvent(r->fd, EVFILT_READ, EV_ADD, 0, 0);
				r->watchRead = YES;
			}
		} else if (watchRead < 0) {
			if (r->watchRead) {
				watchEvent(r->fd, EVFILT_READ, EV_DELETE, 0, 0);
				r->watchRead = NO;
			}
		}

		if (watchWrite > 0) {
			if (!r->watchWrite) {
				watchEvent(r->fd, EVFILT_WRITE, EV_ADD, 0, 0);
				r->watchWrite = YES;
			}
		} else if (watchWrite < 0) {
			if (r->watchWrite) {
				watchEvent(r->fd, EVFILT_WRITE, EV_DELETE, 0, 0);
				r->watchWrite = NO;
			}
		}
#elif MODEL == MODEL_EPOLL
		{
			unsigned oldFlag = 0, newFlag = 0;
			if (r->watchRead) oldFlag = EPOLLIN;
			if (r->watchWrite) oldFlag |= EPOLLOUT;

			if (watchRead) r->watchRead = watchRead > 0 ? YES : NO;
			if (watchWrite) r->watchWrite = watchWrite > 0 ? YES : NO;

			if (r->watchRead) newFlag = EPOLLIN;
			if (r->watchWrite) newFlag |= EPOLLOUT;

			if (oldFlag != newFlag)
				watchEvent(r->fd, newFlag, EPOLL_CTL_MOD);
		}
#elif MODEL == MODEL_POLL
		if (watchRead) r->watchRead = watchRead > 0 ? YES : NO;
		if (watchWrite) r->watchWrite = watchWrite > 0 ? YES : NO;
#endif //MODEL

		r->last = now + config->timeout;
		if (requests != r) {
			if (r->next) r->next->prev = r->prev;
			else tail = r->prev;
			r->prev->next = r->next;
			r->next = requests;
			r->prev = NULL;
			requests->prev = r;
			requests = r;
		}
	}
}

#if MODEL == MODEL_POLL
int
fillProxyPoll(struct pollfd* pollfds)
{
	ProxyRequest* r;
	struct pollfd* pollfd;
	if (!nRequests) return 0;

	for (pollfd = pollfds, r = requests; r; r = r->next) {
		short ev = r->watchRead ? POLLIN : r->watchWrite ? POLLOUT : 0;
		if (ev) {
			pollfd->events = ev;
			pollfd->fd = r->fd;
			pollfd++;
		}
	}
	return pollfd - pollfds;

}
#endif

#endif /* SUPPORT_PROXY */
